"""
opencadd.databases.klifs.schema

Defines schema used across the klifs module.
"""

import pandas as pd

DATAFRAME_COLUMNS = {
    "kinase_groups": [
        ("kinase.group", "string"),
    ],
    "kinase_families": [
        ("kinase.family", "string"),
    ],
    "kinases_all": [
        ("kinase.klifs_id", "int32"),
        ("kinase.hgnc_name", "string"),  # TODO except for kinase KLIFS IDs: 529, 530
        ("kinase.full_name", "string"),
        ("species.klifs", "string"),
    ],
    "kinases": [
        ("kinase.klifs_id", "int32"),
        ("kinase.klifs_name", "string"),  # TODO where from?
        ("kinase.hgnc_name", "string"),
        ("kinase.family", "string"),
        ("kinase.group", "string"),
        ("kinase.class", "string"),  # TODO where from?
        ("species.klifs", "string"),
        ("kinase.full_name", "string"),
        ("kinase.uniprot", "string"),
        ("kinase.iuphar", "string"),
        ("kinase.pocket", "string"),
    ],
    "ligands": [
        ("ligand.klifs_id", "Int32"),  # TODO use int32 when ligand ID avail. locally
        ("ligand.expo_id", "string"),
        ("ligand.name", "string"),
        ("ligand.smiles", "string"),
        ("ligand.inchikey", "string"),
    ],
    "structures": [
        ("structure.klifs_id", "int32"),
        ("structure.pdb_id", "string"),
        ("structure.alternate_model", "string"),
        ("structure.chain", "string"),
        ("species.klifs", "string"),
        ("kinase.klifs_id", "int32"),
        ("kinase.klifs_name", "string"),  # TODO where from?
        # "kinase.names",  # Excluded, otherwise operations like drop_duplicates() do not work
        ("kinase.family", "string"),
        ("kinase.group", "string"),
        ("structure.pocket", "string"),
        ("ligand.expo_id", "string"),
        ("ligand_allosteric.expo_id", "string"),
        ("ligand.name", "string"),
        ("ligand_allosteric.name", "string"),
        ("structure.dfg", "string"),
        ("structure.ac_helix", "string"),
        ("structure.resolution", "float32"),
        ("structure.qualityscore", "float32"),
        ("structure.missing_residues", "int32"),
        ("structure.missing_atoms", "int32"),
        ("structure.rmsd1", "float32"),
        ("structure.rmsd2", "float32"),
        ("structure.front", "boolean"),
        ("structure.gate", "boolean"),
        ("structure.back", "boolean"),
        ("structure.fp_i", "boolean"),
        ("structure.fp_ii", "boolean"),
        ("structure.bp_i_a", "boolean"),
        ("structure.bp_i_b", "boolean"),
        ("structure.bp_ii_in", "boolean"),
        ("structure.bp_ii_a_in", "boolean"),
        ("structure.bp_ii_b_in", "boolean"),
        ("structure.bp_ii_out", "boolean"),
        ("structure.bp_ii_b", "boolean"),
        ("structure.bp_iii", "boolean"),
        ("structure.bp_iv", "boolean"),
        ("structure.bp_v", "boolean"),
        ("structure.grich_distance", "float32"),
        ("structure.grich_angle", "float32"),
        ("structure.grich_rotation", "float32"),
        ("structure.filepath", "string"),
    ],
    "bioactivities": [
        # TODO in the future: "kinase.klifs_id"  # Add if added to KLIFS API?
        ("kinase.pref_name", "string"),
        ("kinase.uniprot", "string"),
        # TODO in the future: "ligand.klifs_id"  # Add if added to KLIFS API?
        ("ligand.bioactivity_standard_type", "string"),
        ("ligand.bioactivity_standard_relation", "string"),
        ("ligand.bioactivity_standard_value", "float32"),
        ("ligand.bioactivity_standard_units", "string"),
        ("ligand.bioactivity_pchembl_value", "float32"),
        ("species.chembl", "string"),
    ],
    "interactions": [
        ("structure.klifs_id", "int32"),
        ("interaction.fingerprint", "string"),
    ],
    "interaction_types": [
        ("interaction.id", "int32"),
        ("interaction.name", "string"),
    ],
    "pockets": [
        ("residue.klifs_id", "int32"),
        ("residue.id", "string"),
        ("residue.klifs_region_id", "string"),
        ("residue.klifs_region", "string"),
        ("residue.klifs_color", "string"),
    ],
    "coordinates": [
        ("atom.id", "int32"),
        ("atom.name", "string"),
        ("atom.x", "float32"),
        ("atom.y", "float32"),
        ("atom.z", "float32"),
        ("residue.id", "string"),  # TODO: int32?
        ("residue.name", "string"),
        ("residue.klifs_id", "int32"),
        ("residue.klifs_region_id", "string"),
        ("residue.klifs_region", "string"),
        ("residue.klifs_color", "string"),
    ],
}

LOCAL_COLUMNS_MAPPING = {
    "klifs_export": {
        "NAME": "kinase.names",  # HGNC and KLIFS (?) name  TODO where is KLIFS name from?
        "FAMILY": "kinase.family",
        "GROUPS": "kinase.group",
        "PDB": "structure.pdb_id",
        "CHAIN": "structure.chain",
        "ALTERNATE_MODEL": "structure.alternate_model",
        "SPECIES": "species.klifs",
        "LIGAND": "ligand.name",
        "PDB_IDENTIFIER": "ligand.expo_id",
        "ALLOSTERIC_NAME": "ligand_allosteric.name",
        "ALLOSTERIC_PDB": "ligand_allosteric.expo_id",
        "DFG": "structure.dfg",
        "AC_HELIX": "structure.ac_helix",
    },
    "klifs_overview": {
        "species": "species.klifs",
        "kinase": "kinase.klifs_name",
        "pdb": "structure.pdb_id",
        "alt": "structure.alternate_model",
        "chain": "structure.chain",
        "orthosteric_PDB": "ligand.expo_id",
        "allosteric_PDB": "ligand_allosteric.expo_id",
        "rmsd1": "structure.rmsd1",
        "rmsd2": "structure.rmsd2",
        "qualityscore": "structure.qualityscore",
        "pocket": "structure.pocket",
        "resolution": "structure.resolution",
        "missing_residues": "structure.missing_residues",
        "missing_atoms": "structure.missing_atoms",
        "full_ifp": "interaction.fingerprint",
        "fp_i": "structure.fp_i",
        "fp_ii": "structure.fp_ii",
        "bp_i_a": "structure.bp_i_a",
        "bp_i_b": "structure.bp_i_b",
        "bp_ii_in": "structure.bp_ii_in",
        "bp_ii_a_in": "structure.bp_ii_a_in",
        "bp_ii_b_in": "structure.bp_ii_b_in",
        "bp_ii_out": "structure.bp_ii_out",
        "bp_ii_b": "structure.bp_ii_b",
        "bp_iii": "structure.bp_iii",
        "bp_iv": "structure.bp_iv",
        "bp_v": "structure.bp_v",
    },
}

REMOTE_COLUMNS_MAPPING = {
    # Information.get_kinase_names()
    "kinases_all": {
        "kinase_ID": "kinase.klifs_id",
        "name": "kinase.hgnc_name",
        "full_name": "kinase.full_name",
        "species": "species.klifs",
    },
    # Information.get_kinase_information()
    "kinases": {
        "kinase_ID": "kinase.klifs_id",
        "name": "kinase.klifs_name",
        "HGNC": "kinase.hgnc_name",
        "family": "kinase.family",
        "group": "kinase.group",
        "kinase_class": "kinase.class",
        "species": "species.klifs",
        "full_name": "kinase.full_name",
        "uniprot": "kinase.uniprot",
        "iuphar": "kinase.iuphar",
        "pocket": "kinase.pocket",
    },
    # Ligands.get_ligands_list
    "ligands": {
        "ligand_ID": "ligand.klifs_id",
        "PDB-code": "ligand.expo_id",
        "Name": "ligand.name",
        "SMILES": "ligand.smiles",
        "InChIKey": "ligand.inchikey",
    },
    # Structures.get_structure_list()
    # Structures.get_structure_lists()
    "structures": {
        "structure_ID": "structure.klifs_id",
        "kinase": "kinase.klifs_name",
        "species": "species.klifs",
        "kinase_ID": "kinase.klifs_id",
        "pdb": "structure.pdb_id",
        "alt": "structure.alternate_model",
        "chain": "structure.chain",
        "rmsd1": "structure.rmsd1",
        "rmsd2": "structure.rmsd2",
        "pocket": "structure.pocket",
        "resolution": "structure.resolution",
        "quality_score": "structure.qualityscore",
        "missing_residues": "structure.missing_residues",
        "missing_atoms": "structure.missing_atoms",
        "ligand": "ligand.expo_id",
        "allosteric_ligand": "ligand_allosteric.expo_id",
        "DFG": "structure.dfg",
        "aC_helix": "structure.ac_helix",
        "Grich_distance": "structure.grich_distance",
        "Grich_angle": "structure.grich_angle",
        "Grich_rotation": "structure.grich_rotation",
        "front": "structure.front",
        "gate": "structure.gate",
        "back": "structure.back",
        "fp_I": "structure.fp_i",
        "fp_II": "structure.fp_ii",
        "bp_I_A": "structure.bp_i_a",
        "bp_I_B": "structure.bp_i_b",
        "bp_II_in": "structure.bp_ii_in",
        "bp_II_A_in": "structure.bp_ii_a_in",
        "bp_II_B_in": "structure.bp_ii_b_in",
        "bp_II_out": "structure.bp_ii_out",
        "bp_II_B": "structure.bp_ii_b",
        "bp_III": "structure.bp_iii",
        "bp_IV": "structure.bp_iv",
        "bp_V": "structure.bp_v",
    },
    # Ligands.get_bioactivity_list_id()
    "bioactivities": {
        "pref_name": "kinase.pref_name",
        "accession": "kinase.uniprot",
        "organism": "species.chembl",
        "standard_type": "ligand.bioactivity_standard_type",
        "standard_relation": "ligand.bioactivity_standard_relation",
        "standard_value": "ligand.bioactivity_standard_value",
        "standard_units": "ligand.bioactivity_standard_units",
        "pchembl_value": "ligand.bioactivity_pchembl_value",
    },
    # Interactions.get_interactions_get_IFP()
    "interactions": {
        "structure_ID": "structure.klifs_id",
        "IFP": "interaction.fingerprint",
    },
    # Interactions.get_interactions_get_types()
    "interaction_types": {
        "position": "interaction.id",
        "name": "interaction.name",
    },
    # Interactions.get_interactions_match_residues()
    "pockets": {
        "index": "residue.klifs_id",
        "Xray_position": "residue.id",
        "KLIFS_position": "residue.klifs_region_id",
    },
}

POCKET_KLIFS_REGIONS = [
    (1, "I"),
    (2, "I"),
    (3, "I"),
    (4, "g.l"),
    (5, "g.l"),
    (6, "g.l"),
    (7, "g.l"),
    (8, "g.l"),
    (9, "g.l"),
    (10, "II"),
    (11, "II"),
    (12, "II"),
    (13, "II"),
    (14, "III"),
    (15, "III"),
    (16, "III"),
    (17, "III"),
    (18, "III"),
    (19, "III"),
    (20, "αC"),
    (21, "αC"),
    (22, "αC"),
    (23, "αC"),
    (24, "αC"),
    (25, "αC"),
    (26, "αC"),
    (27, "αC"),
    (28, "αC"),
    (29, "αC"),
    (30, "αC"),
    (31, "b.l"),
    (32, "b.l"),
    (33, "b.l"),
    (34, "b.l"),
    (35, "b.l"),
    (36, "b.l"),
    (37, "b.l"),
    (38, "IV"),
    (39, "IV"),
    (40, "IV"),
    (41, "IV"),
    (42, "V"),
    (43, "V"),
    (44, "V"),
    (45, "GK"),
    (46, "hinge"),
    (47, "hinge"),
    (48, "hinge"),
    (49, "linker"),
    (50, "linker"),
    (51, "linker"),
    (52, "linker"),
    (53, "αD"),
    (54, "αD"),
    (55, "αD"),
    (56, "αD"),
    (57, "αD"),
    (58, "αD"),
    (59, "αD"),
    (60, "αE"),
    (61, "αE"),
    (62, "αE"),
    (63, "αE"),
    (64, "αE"),
    (65, "VI"),
    (66, "VI"),
    (67, "VI"),
    (68, "c.l"),
    (69, "c.l"),
    (70, "c.l"),
    (71, "c.l"),
    (72, "c.l"),
    (73, "c.l"),
    (74, "c.l"),
    (75, "c.l"),
    (76, "VII"),
    (77, "VII"),
    (78, "VII"),
    (79, "VIII"),
    (80, "xDFG"),
    (81, "xDFG"),
    (82, "xDFG"),
    (83, "xDFG"),
    (84, "a.l"),
    (85, "a.l"),
]
POCKET_KLIFS_REGIONS = pd.DataFrame(
    [
        (klifs_id, klifs_region, ".".join([klifs_region, str(klifs_id)]))
        for (klifs_id, klifs_region) in POCKET_KLIFS_REGIONS
    ],
    columns=["residue.klifs_id", "residue.klifs_region", "residue.klifs_region_id"],
)

POCKET_KLIFS_REGION_COLORS = {
    "I": "khaki",
    "g.l": "green",
    "II": "khaki",
    "III": "khaki",
    "αC": "red",
    "b.l": "green",
    "IV": "khaki",
    "V": "khaki",
    "GK": "orange",
    "hinge": "magenta",
    "linker": "cyan",
    "αD": "red",
    "αE": "red",
    "VI": "khaki",
    "c.l": "darkorange",
    "VII": "khaki",
    "VIII": "khaki",
    "xDFG": "cornflowerblue",
    "a.l": "cornflowerblue",
}
